package com.tigerlogic.omnis.jsclient.android;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserFactory;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.res.AssetManager;
import android.content.res.Configuration;
import android.graphics.Bitmap;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.Window;
import android.view.WindowManager;
import android.webkit.JsResult;
import android.webkit.WebChromeClient;
import android.webkit.WebSettings;
import android.webkit.WebView;
import android.webkit.WebViewClient;
import android.widget.EditText;
import android.widget.ProgressBar;
import android.widget.Toast;

public class OmnisAndroidClient extends Activity {
		static private final String TAG = "OmnisAndroidClient";
		// Preference keys
		static private final String kPrefSaved = "saved";
		static private final String kPrefConfigUrl = "configurl";
		static private final String kPrefTestUrl = "testurl";		
		static private final String kPrefHasTitle = "title";
		static private final String kPrefMaintainAspectRatio = "mar";
		static private final String kPrefScale = "scale";
		static private final String kPrefFloat = "float";
		static private final String kPrefCanScrollH = "hs";
		static private final String kPrefCanScrollV = "vs";
		static private final String kPrefEnableTestForm = "etf";
		static private final String kPrefTestFormServerAndPort = "sap";
		static private final String kPrefStandardMenu = "sm";
		
		// Start of properties to be set via configuration
		private String mCurrentUrl;				// Current displayed URL
		private String mTestUrl;
		private boolean mHasTitle = false;
		private boolean mMaintainAspectRatio = false;
		// If scale is true, controls do not float, so mFloatControls is ignored
		private boolean mScale = false;
		private boolean mFloatControls = false;
		private boolean mCanScrollH = false;
		private boolean mCanScrollV = false;
		
		private boolean mEnableTestForm = false;
		private String mTestFormServerAndPort;
		
		private boolean mStandardMenu = true;
		// End of properties to be set via configuration
		private ProgressIndicator progressIndicator;
		private WebView mWebView;
		private OmnisAndroidLinearLayout mMainContainer;
		public static int smOrientation;
		private OmnisAndroidTestFormThread mTestFormThread;
		private Handler mHandler = new Handler();
		
		private boolean mWindowHasTitle;	// Real state of the window's has title flag
		
    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        
        if (!readInitialConfiguration()) {
         	finish();
         	return;
        }  
        // Can only remove title here, before setting content view
        if (!mHasTitle)
        		requestWindowFeature(Window.FEATURE_NO_TITLE);
        mWindowHasTitle = mHasTitle;
        // Also remove top status bar
        getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN, WindowManager.LayoutParams.FLAG_FULLSCREEN);
        setContentView(R.layout.main);        
        
        progressIndicator = new ProgressIndicator(); //Instantiate progress bar
        
        mMainContainer = (OmnisAndroidLinearLayout) findViewById(R.id.maincontainer);
        mWebView = (WebView) findViewById(R.id.webview);
        mWebView.clearCache(true);	// Clear entire cache for now - it does not seem to pick up changes to scripts too well without doing this
        
        WebSettings webSettings = mWebView.getSettings();
        webSettings.setJavaScriptEnabled(true);
        webSettings.setDomStorageEnabled(true);	// Enables local storage
        webSettings.setDatabasePath("/data/data/com.tigerlogic.omnis.jsclient.android/databases/");	// Needed to make local storage work correctly
        webSettings.setSupportZoom(true);
        
        mWebView.setWebViewClient(new WebViewClient() {
          @Override
          public void onReceivedError(WebView view, int errorCode, String description, String failingUrl) {
          	Log.d(TAG, description);
          	Toast.makeText(OmnisAndroidClient.this, description, Toast.LENGTH_LONG).show();
          	super.onReceivedError(view, errorCode, description, failingUrl);
          }
          
          @Override
        	public void onPageStarted(WebView view, String url, Bitmap favicon) {
        		progressIndicator.displayProgress();
        		super.onPageStarted(view, url, favicon);
        	}
          
          @Override
        	public void onPageFinished(WebView view, String url) {
        		progressIndicator.hideProgress();
        		super.onPageFinished(view, url);
        	}
        });
        
        mWebView.setWebChromeClient(new WebChromeClient() {
        	public boolean onJsAlert (WebView view, String url, String message, final JsResult result) {
        			new AlertDialog.Builder(OmnisAndroidClient.this)  
        				.setTitle("")  
        				.setMessage(message)  
        				.setPositiveButton(android.R.string.ok,  
                    new AlertDialog.OnClickListener()  
                    {  
												@Override
												public void onClick(DialogInterface arg0, int arg1) {
                          result.confirm();  
												}  
                    })  
                .setCancelable(false)  
                .create()  
                .show();  
        			return true;  
        	}
        	
        	@Override
        	public void onConsoleMessage(String message, int lineNumber, String sourceID) {
          	Log.d(TAG, message);
        		Toast.makeText(OmnisAndroidClient.this, message, Toast.LENGTH_LONG).show();
        		super.onConsoleMessage(message, lineNumber, sourceID);
        	}  
        	
        	@Override
        	public boolean onJsTimeout() {
        		String message = "JavaScript execution timeout";
          	Log.d(TAG, message);
        		Toast.makeText(OmnisAndroidClient.this, message, Toast.LENGTH_LONG).show();
        		return super.onJsTimeout();
        	}
        	
        	@Override 
        	public void onProgressChanged(WebView view, int newProgress) {
        		//Update progress bar
        		progressIndicator.updateProgress(newProgress);
        		super.onProgressChanged(view, newProgress);
        	}
        });
        
        smOrientation = getResources().getConfiguration().orientation;
        mWebView.addJavascriptInterface(new OmnisAndroidJSInterface(), "omnisandroid");
        
        if (!mEnableTestForm)
        	doLoad(mCurrentUrl, false);
        else {
        	if (mTestUrl != null && mTestUrl != "") {
        		// Restarting as the title state is changing
        		doLoad(mTestUrl, true);
        	}
        }
    }
    
    @Override
    protected void onStart() {
    	super.onStart();
      checkTestMode();
    }
    
    @Override
    protected void onStop() {
    	doStopTestFormThread();
    	super.onStop();
    }
    
    public boolean onCreateOptionsMenu(Menu menu) {
    		if (!mStandardMenu)
    			return false;
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.menu, menu);
        return true;
    }
    
    @Override
    public boolean onPrepareOptionsMenu(Menu menu) {
    	menu.findItem(R.id.floatcontrols).setChecked(mFloatControls);
    	menu.findItem(R.id.floatcontrols).setIcon(mFloatControls ? android.R.drawable.checkbox_on_background : android.R.drawable.checkbox_off_background);
    	menu.findItem(R.id.scale).setChecked(mScale);
    	menu.findItem(R.id.maintainaspect).setChecked(mMaintainAspectRatio);
    	menu.findItem(R.id.canscrollh).setChecked(mCanScrollH);
    	menu.findItem(R.id.canscrollv).setChecked(mCanScrollV);
    	menu.findItem(R.id.enabletest).setChecked(mEnableTestForm);
    	menu.findItem(R.id.enabletest).setIcon(mEnableTestForm ? android.R.drawable.checkbox_on_background : android.R.drawable.checkbox_off_background);
    	return super.onPrepareOptionsMenu(menu);
    }
    
    public boolean onOptionsItemSelected(MenuItem item) {
      switch (item.getItemId()) {
      	case R.id.reload:
      		doLoad(mCurrentUrl, false);
          return true;
      	case R.id.floatcontrols:
      		mFloatControls = !mFloatControls;
      		break;
      	case R.id.scale:
      		mScale = !mScale;
      		break;
      	case R.id.maintainaspect:
      		mMaintainAspectRatio = !mMaintainAspectRatio;
      		break;
      	case R.id.canscrollh:
      		mCanScrollH = !mCanScrollH;
      		break;
      	case R.id.canscrollv:
      		mCanScrollV = !mCanScrollV;
      		break;
      	case R.id.enabletest:
      		mEnableTestForm = !mEnableTestForm;
      		break;
      	case R.id.editurl:
      		editText(mCurrentUrl, true);
      		break;
      	case R.id.edittestserver:
      		editText(getTestFormServerAndPort(null), false);
      		return true;
      	default:
          return super.onOptionsItemSelected(item);
      }
      // All the items which get here require the preferences to be saved
      savePreferences();
      // Check test mode again
      checkTestMode();
      return true;
    }
    
    public void onConfigurationChanged (Configuration newConfig) {
    	super.onConfigurationChanged(newConfig);
    	smOrientation = newConfig.orientation;
    }
    
    public void editText(final String pCurrentValue, final boolean pURL) {
	    AlertDialog.Builder alert = new AlertDialog.Builder(this);
	
	    alert.setTitle(getString(R.string.app_name));
	    alert.setMessage(getString(pURL ? R.string.editurl : R.string.ipandport));
	
	    // Set an EditText view to get user input 
	    final EditText input = new EditText(this);
	    input.setText(pCurrentValue);
	    alert.setView(input);
	
	    alert.setPositiveButton("Ok", new DialogInterface.OnClickListener() {
	    	public void onClick(DialogInterface dialog, int whichButton) {
	    		String value = input.getText().toString();
	    		if (pURL) {
	    			mCurrentUrl = value;
    				savePreferences();
	    		} else {
	    			if (!value.equals(pCurrentValue)) {
	    				setTestFormServerAndPort(value);
	    				savePreferences();
	    				checkTestMode();
	    			}
	    		}
	      }
	    });
	
	    alert.setNegativeButton("Cancel", new DialogInterface.OnClickListener() {
	      public void onClick(DialogInterface dialog, int whichButton) {
	        // Cancelled.
	      }
	    });
	
	    alert.show();
    }
    
    private boolean readInitialConfiguration() {
    	if (loadPreferences()) {
    		return true;
    	}
    	boolean ok = false;
    	do {
    		AssetManager am = getAssets();
    		if (am != null) {
    			InputStream cis = null;
					try {
						cis = am.open("config.xml");
					} catch (IOException e) {
						e.printStackTrace();
					}
    			if (cis != null) {
    				XmlPullParserFactory factory = null;
						try {
							factory = XmlPullParserFactory.newInstance();
							if (factory != null) {
								factory.setNamespaceAware(false);
								XmlPullParser xpp = factory.newPullParser();
								if (xpp != null) {
									xpp.setInput(cis, null);	// null means detect encoding from input
			            int eventType = xpp.getEventType();
			            boolean done = false;
			            while (eventType != XmlPullParser.END_DOCUMENT && !done) {
		                String name = null;
		                switch (eventType){
		                    case XmlPullParser.START_TAG:
		                        name = xpp.getName();
                            if (name.equalsIgnoreCase("URL")){
                            	mMaintainAspectRatio = getBooleanAttribute(xpp, "maintainaspectratio");
                            	mScale = getBooleanAttribute(xpp, "scale");
                            	mFloatControls = getBooleanAttribute(xpp, "floatcontrols");
                            	mCanScrollH = getBooleanAttribute(xpp, "canhorzscroll");
                            	mCanScrollV = getBooleanAttribute(xpp, "canvertscroll");
                            	mStandardMenu = getBooleanAttribute(xpp, "standardmenu");
                            	mHasTitle = getBooleanAttribute(xpp, "title");
                             	mCurrentUrl = xpp.nextText().trim();
                              ok = true;
                            	done = true;
                            } else if (name.equalsIgnoreCase("TEST")) {
                            	setTestFormServerAndPort(xpp.nextText().trim());
                            	mEnableTestForm = getBooleanAttribute(xpp, "enable");
                            }
                            break;
		                }
		                eventType = xpp.next();
			            }
								}
							}
						}
						catch (Exception e) {
						}
						try {
							cis.close();
						} catch (IOException e) {
						}
    			}
    		}
    	} while (false);
    	if (ok) {
    		savePreferences();
    	}
    	return ok;
    }
    
    private boolean getBooleanAttribute(XmlPullParser xpp, String pAttName) {
    	String value = xpp.getAttributeValue(null, pAttName);
    	if (value == null)
    		return false;
    	if (value.equals("1"))
    		return true;
    	return false;
    }
    
    private boolean loadPreferences() {
    	SharedPreferences prefs = getPreferences(Context.MODE_PRIVATE);
    	if (prefs.getBoolean(kPrefSaved, false)) {
    		mCurrentUrl = prefs.getString(kPrefConfigUrl, "");	
    		mTestUrl = prefs.getString(kPrefTestUrl, "");	
    		mHasTitle = prefs.getBoolean(kPrefHasTitle, false);
    		mMaintainAspectRatio = prefs.getBoolean(kPrefMaintainAspectRatio, true);
    		mScale = prefs.getBoolean(kPrefScale, false);
    		mFloatControls = prefs.getBoolean(kPrefFloat, true);
    		mCanScrollH = prefs.getBoolean(kPrefCanScrollH, false);
    		mCanScrollV = prefs.getBoolean(kPrefCanScrollV, false);
    		mStandardMenu = prefs.getBoolean(kPrefStandardMenu, true);
    		
    		mEnableTestForm = prefs.getBoolean(kPrefEnableTestForm, false);
    		setTestFormServerAndPort(prefs.getString(kPrefTestFormServerAndPort, ""));
    		return true;
    	}
    	return false;
    }
    
    private void savePreferences() {
    	SharedPreferences prefs = getPreferences(Context.MODE_PRIVATE);
    	SharedPreferences.Editor editor = prefs.edit();
    	editor.putString(kPrefConfigUrl, (mCurrentUrl != null) ? mCurrentUrl : "");
    	editor.putString(kPrefTestUrl, (mTestUrl != null) ? mTestUrl : "");
    	editor.putBoolean(kPrefHasTitle, mHasTitle);
    	editor.putBoolean(kPrefMaintainAspectRatio, mMaintainAspectRatio);
    	editor.putBoolean(kPrefScale, mScale);
    	editor.putBoolean(kPrefFloat, mFloatControls);
    	editor.putBoolean(kPrefCanScrollH, mCanScrollH);
    	editor.putBoolean(kPrefCanScrollV, mCanScrollV);
    	editor.putBoolean(kPrefStandardMenu, mStandardMenu);
  		
    	editor.putBoolean(kPrefEnableTestForm, mEnableTestForm);
    	String testFormServerAndPort = getTestFormServerAndPort(null);
    	editor.putString(kPrefTestFormServerAndPort, (testFormServerAndPort != null) ? testFormServerAndPort : "");
    	editor.putBoolean(kPrefSaved, true);
    	editor.commit();
    }
    
    private synchronized String getTestFormServerAndPort(String pOldValue) {
    	if (pOldValue != null && pOldValue.equals(mTestFormServerAndPort)) {
    		return pOldValue;
    	}
    	return new String((mTestFormServerAndPort == null) ? "" : mTestFormServerAndPort);
    }
    
    private synchronized void setTestFormServerAndPort(String pTestFormServerAndPort) {
    	mTestFormServerAndPort = pTestFormServerAndPort;
    }

    private synchronized boolean checkTestMode() {
    	// Check to see if the test mode thread needs to run.
    	// If not, and it is running, stop it
    	// If it does, and it is not running, start it
    	if (!mEnableTestForm) {
    		if (mTestFormThread != null) {
    			mTestFormThread.setStopThread(true);
    			mTestFormThread = null;
    		}
    		return false;
    	}
    	if (mTestFormThread == null) {
    		mTestFormThread = new OmnisAndroidTestFormThread();
    		mTestFormThread.start();
    	}
    	return true;
    }
    
    private synchronized void doStopTestFormThread() {
    	if (mTestFormThread != null) {
    		mTestFormThread.setStopThread(true);
    		mTestFormThread = null;
    	}
    }
    
    private void doLoad(String pURL, boolean pTestForm) {
    	if (mHasTitle != mWindowHasTitle) {
    		// Need to restart
    		savePreferences();
    		
    		Intent intent = getIntent();
    		finish();
    		startActivity(intent);
    		return;
    	}
      if (pURL != null && pURL.length() > 0) {
      	mWebView.loadUrl(pURL);
      	if (pTestForm) {
      		mTestUrl = "";
      		savePreferences();
      	}
      }
    }
    
    class OmnisAndroidJSInterface {
    	
    	OmnisAndroidJSInterface() {
    	}
    	
    	public int getAvailScreenWidth() {
    		OmnisAndroidClient oac = OmnisAndroidClient.this;
    		return (OmnisAndroidClient.smOrientation == Configuration.ORIENTATION_PORTRAIT) ? oac.mMainContainer.mPortraitWidth : oac.mMainContainer.mLandscapeWidth;
    	}
    	
    	public int getAvailScreenHeight() {
     		OmnisAndroidClient oac = OmnisAndroidClient.this;
     	  return (OmnisAndroidClient.smOrientation == Configuration.ORIENTATION_PORTRAIT) ? oac.mMainContainer.mPortraitHeight : oac.mMainContainer.mLandscapeHeight;
    	}
    	
    	public boolean getMaintainAspectRatio() {
     		return OmnisAndroidClient.this.mMaintainAspectRatio;
    	}
    	
    	public boolean getScale() {
     		return OmnisAndroidClient.this.mScale;
    	}
    	
    	public boolean getFloatControls() {
    		return OmnisAndroidClient.this.mFloatControls;
    	}

    	public boolean getCanScrollH() {
    		return OmnisAndroidClient.this.mCanScrollH;
    	}

    	public boolean getCanScrollV() {
    		return OmnisAndroidClient.this.mCanScrollV;
    	}
    	
    	public boolean isLandscape() {
    		return OmnisAndroidClient.smOrientation == Configuration.ORIENTATION_LANDSCAPE; 
    	}
    }
    
    class OmnisAndroidTestFormThread extends Thread {
    	boolean mStopThread = false;
  		OmnisAndroidSplitAddress mAddressAndPort = new OmnisAndroidSplitAddress();
  		Socket mSocket = null;
  		InputStream mInputStream = null;
  		int mFileNameLength = 0;
    	
    	public synchronized void setStopThread(boolean pStopThread) {
    		mStopThread = pStopThread;
    	}
    	
    	public synchronized boolean getStopThread() {
    		return mStopThread;
    	}
    	
    	private void closeSocket() {
    		try {
    			if (mSocket != null)
    				mSocket.close();
    		} catch (IOException e) {
    		}
    		mSocket = null;
    		mInputStream = null;
    		mFileNameLength = 0;
    	}
    	
    	public void run() {
    		String testFormServerAndPort = null;
    		while (!getStopThread()) {
    			testFormServerAndPort = getTestFormServerAndPort(testFormServerAndPort);
    			boolean newAddress = mAddressAndPort.split(testFormServerAndPort);
    			if (mSocket != null) {
    				if (newAddress) {
    					// Address has changed - we need a new socket
    					closeSocket();
    				}
    			}
    			if (mSocket == null && mAddressAndPort.mIpAddress != null && mAddressAndPort.mPort != 0) {
  					try {
      				mSocket = new Socket(mAddressAndPort.mIpAddress, mAddressAndPort.mPort);
      				byte[] connectMessage = new byte[23];
      				connectMessage[0] = 0; 	// Start overall message length
      				connectMessage[1] = 0;
      				connectMessage[2] = 0;
      				connectMessage[3] = 23;	// End overall message length
      				connectMessage[4] = 'O';
      				connectMessage[5] = 'M';
      				connectMessage[6] = 'N';
      				connectMessage[7] = 'I';
      				connectMessage[8] = 'S';
      				connectMessage[9] = '_';
      				connectMessage[10] = 'O';
      				connectMessage[11] = 'R';
      				connectMessage[12] = 'F';
      				connectMessage[13] = 'C';
      				connectMessage[14] = 0;
      				connectMessage[15] = 1;	// Version
      				connectMessage[16] = 0;
      				connectMessage[17] = 40;// JavaScript test form connect message
      				connectMessage[18] = 0; // Start client IP address - zero for a local client
      				connectMessage[19] = 0;
      				connectMessage[20] = 0;
      				connectMessage[21] = 0; // End client IP address - zero for a local client
      				connectMessage[22] = 0;	// Length of Omnis server address - zero
  					
      				OutputStream os = mSocket.getOutputStream();
      				os.write(connectMessage);
      				os.flush();
      				
      				mInputStream = mSocket.getInputStream();
  					} catch (IOException e) {
  						closeSocket();
  					} 					
    			}
    			if (mSocket != null) {
    				// Try to receive a new HTML file name from the server
    				try {
    					if (mFileNameLength == 0) {
    						// First 4 bytes of message are the total length
    						if (mInputStream.available() >= 4) {
    							// Read length component.  This is followed by a further 14 bytes we can ignore (the web client message header), and then the title indicator and file name
    							byte[] cl = new byte[4];
    							mInputStream.read(cl);
    							mFileNameLength = (((int) cl[0]) << 24) + (((int) cl[1]) << 16) + (((int) cl[2]) << 8) + ((int) cl[3]);
    							if (mFileNameLength <= 18) {
    								// Should not get to here
    								closeSocket();
    							}
    							mFileNameLength -= 4;	
    						}
    					}
    					if (mInputStream != null && mFileNameLength > 0 && mInputStream.available() >= mFileNameLength) {
    						// The new config is the new HTML file name to use, with a prefix of '0' if no title is required, or '1' if a title is required
    						// Read and discard the 14 bytes of the header
    						byte[] header = new byte[14];
    						mInputStream.read(header);
    						mFileNameLength -= 14;
    						// Read the title indicator
    						byte[] titleIndicator = new byte[1];
    						mInputStream.read(titleIndicator);
    						mFileNameLength -= 1;
    						// Read the actual file name
    						byte[] newFileNameBytes = new byte[mFileNameLength];
    						mFileNameLength = 0;
    						mInputStream.read(newFileNameBytes);
    						
    						final String newFileName = new String(newFileNameBytes);	// Converts using default charset, UTF-8 on Android
    						mHasTitle = (titleIndicator[0] == '0') ? false : true;
    						mHandler.post(new Runnable() {	
    							// This will run in the handler in the user interface thread
									@Override
									public void run() {
										mTestUrl = "http://" + mAddressAndPort.mServerAddressAndPort + "/jschtml/" + newFileName;
										doLoad(mTestUrl, true);
									}
								});
    					}
						} catch (IOException e) {
  						closeSocket();
						}
    			}
    			// Suspend for 250ms before trying again
    			try {
						Thread.sleep(250);
					} catch (InterruptedException e) {
					}
    		}
    		// Make sure socket is closed before exiting thread
    		closeSocket();
    	}
    }
    
    class OmnisAndroidSplitAddress {
    	String  mServerAddressAndPort;
    	String 	mIpAddress;
    	int			mPort;
    	
    	OmnisAndroidSplitAddress() {
    	}
    	
    	boolean split(String pServerAddressAndPort) {
    		// Returns true for a new address
    		if (mServerAddressAndPort != null) {
    			if (mServerAddressAndPort.equals(pServerAddressAndPort)) {
    				return false;
    			}
    		}
    		mServerAddressAndPort = new String(pServerAddressAndPort);
    		mIpAddress = null;
    		mPort = 0;
    		int colonPos = mServerAddressAndPort.indexOf(':');
    		if (colonPos != -1) {
    			mIpAddress = mServerAddressAndPort.substring(0, colonPos);
    			mPort = Integer.valueOf(mServerAddressAndPort.substring(colonPos + 1));
    		}
    		return true;
    	}
    }
    
    class ProgressIndicator {
    	ProgressBar progBar;
    	
    	ProgressIndicator() {
    		progBar = (ProgressBar) findViewById(R.id.progressBar);
    	}
    	
    	public void displayProgress() {
    		progBar.setProgress(0);
    		progBar.setVisibility(View.VISIBLE);
    	}
    	
    	public void updateProgress(int amount) {
    		progBar.setProgress(amount);
    		if (amount == 100) {
    			hideProgress();
    		}
    	}
    	
    	public void hideProgress() {
    		progBar.setVisibility(View.GONE);
    	}
    }
}
